#include "stdafx.h"
#include "Cchess_1.h"
#include "Cchess_2.h"
#include "Cchess_4.h"
#include "Cchess_5.h"
#include "Cchess_6.h" 


// Zobrist hast tables 
// L'utilisation de xor permet de faire un hashage incrmental


/*
  Functions dealing with the standard hash table and the pawn hash tables.
  The standard hash table is divided into two pieces.  The lower half uses
  a "smart" replace strategy, and the upper half uses an always replace
  strategy.
  The pawn is very much like the standard hash - but simpler.  Only
  the hash key and score is stored.  The pawn table is always replace.
*/


/* Function for or'ing two hash codes */
#if INCREMENTAL_HASH==1
inline void or(H_CODE &A, H_CODE B)
{
//A.address=A.address^B.address;
//A.key=A.key^B.key;
A.address^=B.address;
A.key^=B.key;

//A.key2=A.key2^B.key2;
return;
}
#else
//inline H_CODE or(H_CODE &A, H_CODE &B)
inline void or(H_CODE &A, H_CODE &B)
{
 //H_CODE temp_code;
 //temp_code.address = A.address^B.address;
 //temp_code.key = A.key^B.key;
 //return temp_code;
 //A.address = A.address^B.address;
 //A.key = A.key^B.key;
 A.address^=B.address;
 A.key^=B.key;
 //A.key2 = A.key2^B.key2;
 return;
}
#endif


/* Function to set the hash table size */
void set_hash_tables(short Hash_Mbytes, short Pawn_Mbytes)
{
  unsigned long max_tab_entries;
  unsigned long max_pawn_entries;


  // Initialisations
  xhash_tab_entry_size = sizeof(HASH_REC);
  xhash_pawn_entry_size= sizeof(HASH_PAWN_REC);
  xhash_tab_entries=0;
  xhash_pawn_entries=0;
  xhash_succes_creation=false;
  

  //if(Mbytes >= 8) Max_pawn_size = 2000000/(pawn_entry_size);
  //else Max_pawn_size = Mbytes*1024000/(4*pawn_entry_size);
  //PAWN_SIZE = 2;
  //while(2*PAWN_SIZE < Max_pawn_size) { PAWN_SIZE *= 2; }

 
  max_tab_entries = Hash_Mbytes*1048576/xhash_tab_entry_size;
  xhash_tab_entries = 2;
  while(2*xhash_tab_entries <= max_tab_entries) { xhash_tab_entries *= 2; }

  max_pawn_entries = Pawn_Mbytes*1048576/xhash_pawn_entry_size;
  xhash_pawn_entries = 2;
  while(2*xhash_pawn_entries <= max_pawn_entries) { xhash_pawn_entries *= 2; }

  // 1Ko = 1024 
  // 16 records par Ko
  // son on veut un nombre d'entre multiple de 2
  // Cela donne les possibilits suivantes en nombre de Ko
  // 512Ko - 1024Ko ( 1Mo )
  
  xhash_tab_entries_div2 = (xhash_tab_entries>>1);
  xhash_tab_entries_div2_1 = (xhash_tab_entries>>1)-1;
  xhash_pawn_entries_div2 = (xhash_pawn_entries)>>1;
  xhash_pawn_entries_div2_1 = (xhash_pawn_entries>>1)-1;


  close_hash();
  open_hash();

  if (xuci_mode)
  {
	sprintf(xbuffer,"Hash table size : %d",(xhash_tab_entries*xhash_tab_entry_size)>>20);
	display_info(xbuffer);
	sprintf(xbuffer,"Hash pawn-table size : %d",(xhash_pawn_entries*xhash_pawn_entry_size )>>20);
	display_info(xbuffer);
  } // END_MODE;


}

/* Function with allocates space for the hash tables and generates
   the hash codes */
void open_hash()
{

#if NO_PUT_HASH==1 && NO_GET_HASH==1
 char cc;
 cc=0;
 return;
#endif

 start_code();
 // Amusant :
 //hash_table = new HASH_REC[TAB_SIZE];
 //pawn_table = new PAWN_REC[PAWN_SIZE];

 //if (xconsole_mode) { printf("Size of HASH_REC : %d", sizeof(HASH_REC)); } // END_MODE;
xhash_table=NULL;
if (xhash_tab_entries>10) // mieux que 20
{
xhash_table=(HASH_REC*)calloc((size_t) xhash_tab_entries,sizeof(HASH_REC));
}

#if USE_PAWN_HASH_TABLE==1
xhash_pawn_table=(HASH_PAWN_REC*)calloc((size_t) xhash_pawn_entries,sizeof(HASH_PAWN_REC));
#endif

 if ( xhash_table!=NULL )
	 xhash_succes_creation=true;
 else 
	 xhash_succes_creation=false;
 
} 
// Prcaution
//reset_hash_tables();

#if USE_HASH_TABLE==1
void reset_hash_table(void)
{

if ( !xhash_succes_creation ) return;


#if NO_RESET_HT_SEARCH_ENTRIES==0 
memset(xhash_table,0,xhash_tab_entries*sizeof(HASH_REC));
#else
unsigned long index;
HASH_REC  *h;
hptr=xhash_table;
for (index=0;index<xhash_tab_entries;index++,h++)
{
#if NO_RESET_HT_SEARCH_ENTRIES==1
if ((h->flag>>2)>=2) h->key=0;
#endif
#if RESET_MATE_VALUES==1
if		(hptr->score> MATE_VALUE_THRESHOLD) h->key=0;
else if (hptr->score<-MATE_VALUE_THRESHOLD) h->key=0;
#endif
}
#endif
}

/*
void reset_hash_indice_generation()
{
unsigned long index;
HASH_REC  *h;
h=xhash_table;
if (xtempsprevu[1-xwtm]>3) // Au moins 3 secondes de rflexion
{
for (index=0;index<xhash_tab_entries_div2;index++,h++)
{
	h->hash_indice_generation=-1;
	//h->signature_nbcoups=-1;
}
}
// else reset_hash_tables(); prendrait trop de temps
}
*/

/*
void clean_hash_table(unsigned __int16 lv1_move)
{
unsigned long index;
HASH_REC  *h;
h=xhash_table;
if (xtempsprevu[1-xwtm]>3) // Au moins 3 secondes de rflexion
{
for (index=0;index<xhash_tab_entries_div2;index++,h++)
{
	if (h->lv1_move!=lv1_move) h->key=0;
}
}
}
*/

#endif

/* function to close the hash table */
void close_hash()
{
 

#if NO_PUT_HASH==1 && NO_GET_HASH==1
 char cc;
 cc=0;
 return;
#endif

 // Amusant :
 //if (hash_table!=NULL) delete [] hash_table;
 //if (pawn_table!=NULL) delete [] pawn_table;

 if (xhash_table != NULL) 
	 free(xhash_table);
 if (xhash_pawn_table != NULL ) 
	 free(xhash_pawn_table);

}




#if INCREMENTAL_HASH==1
void put_pawn_hash(H_CODE &h_key)
{
#else
void put_pawn_hash(H_CODE &h_key)
{
#endif
 HASH_PAWN_REC *h, *r;
 STATEVALUES *s;
 char x;
 char cc;

 // Initialiser s
 s=&xstates[xlevel];

 // Find location in table
 h = xhash_pawn_table + (((long)h_key.address)&(xhash_pawn_entries_div2_1));

 // Toujours copier en premire partie 
 // Ne recopier en seconde partie 
 // l'ancienne entre que s'il y a collision
 if (h->key!=0)
 {
 if (h->key!=(long)h_key.key)
 { 
 //xnb_put_pawn_hash_collisions++;
 r = h + xhash_pawn_entries_div2; 
 (*r) = (*h);  // copy entry to upper half of table
 }
 }
 xnb_put_pawn_hash++; r = h + xhash_pawn_entries_div2; 
 h->key = (long) h_key.key;
 h->pawnstruct = s->pawnstruct;
 h->passed[1] = s->passed[1];
 h->passed[0] = s->passed[0];
 h->passed_maxpos[1] = s->passed_maxpos[1];
 h->passed_maxpos[0] = s->passed_maxpos[0];
 h->colonnes_ouvertes=0;
 h->roque[1]=s->roque[1];
 h->roque[0]=s->roque[0];
 h->king_safety[1]=s->king_safety[1];
 h->king_safety[0]=s->king_safety[0];

 //for(i=0;i<=1;i++)
 //{
   for(x=0;x<=7;x++)
   {
	   cc=xcol[1][x][statut]; // Statut
	   if (cc==COLONNE_OUVERTE)
	   {
	   	  set_bit_to_1(h->colonnes_ouvertes,x);
	  	  #if DEBUG==1
		  char dd;
		  dd=xcol[0][x][statut];
	      if (cc!=dd)
			 log_error("\nErr H_453");
	      #endif
	   }
	   //else if (cc==COLONNE_SEMI_OUVERTE)
   	   //	  set_bit_to_1(h->colonnes_semi_ouvertes[i],x);
   }
// }


}


/* function to stuff a hash entry into the hash table */
/*
void put_hash(H_CODE h_key, int score, int alpha, int beta,
  int depth, move hmove)
*/

void put_hash(H_CODE &h_key, short score, short depth_moins_level, 
			  char F_alphabeta, MOVE hmove
			  #if HASH_TRACK_PV==1
			  ,char next_node_type
			  #endif
			  )
{
 HASH_REC *h; 
// short depth_moins_level;

#if DEBUG==1
 char cc=0;
 char dd=1;	
 HASH_REC *last_hash_table;
 #endif

 #if NO_PUT_HASH==1
 return;
 #endif





if ( !xhash_succes_creation ) return;


#if DEBUG==1
if (
	
    ( 1==1
	  && h_key.address==1337364217
	  && h_key.key==3320443943 
	  && score==-83
	  //&& (depth>>FD)-xlevel==4
	 )
	)
{
printf("\nTrouv");
//getch(); 
}
#endif


 // Rejeter les scores nuls
 // on choisit plutt de ne pas les stocker
 // Autre option : ne pas stocker les scores nuls 
 // au moment de la rptition
// C'est cette option qui est retenue
//  if (score==0) return;


 // Traitement hmove
 if (!MOVE_IS_FOR_PLAY(hmove)) 
	  hmove=0;
 

/*
// Dbogage
#if DEBUG==1
COINSOLE_MODE;
if (1==2
&& score==-90 
&& (xligne[0].move->dep ==68 && xligne[0].move->arr==35) 
&& (xligne[1].move->dep==63 && xligne[1].move->arr==55) 
&& (xligne[2].move->dep==47 && xligne[2].move->arr==46) 
&& (xligne[3].move->dep==55 && xligne[3].move->arr==43) 
&& (xligne[4].move->dep==28 && xligne[4].move->arr==47) 
&& (xligne[5].move->dep==43 && xligne[5].move->arr==35) 
&& (xligne[6].move->dep==46 && xligne[6].move->arr==35) 
	 // && h_key.key==433601223
	 // && h_key.address==1864182647
	 ) 
 {
	 printf("\nTrouv");
	 getch();
	}
 END_MODE;
 #endif
*/






 // A implmenter : en fait, il faut retirer toutes 
 // les rptitions
 // => il faudrait scanner la suite pour dtecter les rptitions
 // il peut y avoir des rptitions puis une premire fois une suite
 // => la dtection des rptitions ne dtecte alors rien
 // Exemple : Txb2 Cb1 Da4 Cc3 Da5 Rxb2
 // au lieu de Txb2 Rxb2
 // Cela doit tre craint, car la valeur Txb2 Rxb2 Fxc3
 // est crase par une valeur moins pertinente


 // C'est faisable !
 // si on stocke sparment xnbcoups+xlevel
 // et *depth>>FD-xlevel
 // Si deux positions matchent, la diffrence sur 
 // xnbcoups+xlevel donne le nombre de rptitions !!
 // Le coup correspond si aprs dduction des rptitions,
 // la profondeur analyse reste suffisante
 // L'ide est  retenir !!


 /*
 // is this an upper bound, lower bound, or exact score?
 if (score >= beta) { flag = FLAG_B; score = beta; }
 else if (score <= alpha) { flag = FLAG_A; score = alpha;  hmove.t = NOMOVE; }
 else { flag = FLAG_P; }
 */

/*
// Ajustement au niveau
// L'ide est que la valeur mat est stocke  la valeur mat+ply
// et recupre  mat+ply-ply soit mat  ply
// Rutilisation des hash tables
// le mat est enregistr  xlevel==3  une valeur de -MATE_VALUE-3
// il est stock en hash tables pour -MATE-VALUE-3+3==-MATE_VALUE
// Lorsqu'il est retrouv au level 1 aprs que 2 coups ont t jous
// il est rcupr pour -MATE-VALUE-1 soit mat en 1

// Essai : 
// le mat est enregistr  xlevel==3  une valeur de -MATE_VALUE+3
// Lorsqu'il est retrouv au level 1 aprs que 2 coups ont t jous
// il est rcupr pour -MATE-VALUE+3-1 soit mat en 1
if (score>MATE_VALUE_THRESHOLD)
{
	score=inf(MAX_VALUE,score-xlevel);

}
else if (score<-MATE_VALUE_THRESHOLD)
{
	score=sup(-MAX_VALUE,score+xlevel);
}
*/


#if _DEBUG
if (abs(score)>MAX_VALUE)
log_error("H_632",xcompteur);
#endif



// adjust any mate score to be relative to this node.
// if(score >= MATE/2) score += ply;
// else if(score <= -MATE/2) score -= ply;

 // Find location in table
h = xhash_table + (((long)h_key.address)&(xhash_tab_entries_div2_1));

// Use "smart" replace strategy if possible
/*
 if ((h->id != h_id) || (h->depth < depth)
//    || (h->depth == depth && flag != FLAG_A))
*/




// Contrle des adresses pointes
#if DEBUG==1
HASH_REC *r; 
last_hash_table=xhash_table+xhash_tab_entries-1;
r = h + (xhash_tab_entries_div2);
if (h<xhash_table) cc=1;
if (h>last_hash_table) cc=1;
if (r<xhash_table) cc=1;
if (r>last_hash_table) cc=1;
if (cc==1)
{
log_error("\nErreur H_001");
}
#endif


/*
// Effacement slectif des entres anciennes
if (h->age!=xcurrent_hashtable_age)
{
	h->age=xcurrent_hashtable_age;
	if (h->lv1_dep!=xold_lv1_dep)
	{
		goto STORE;
	}
	else if (h->lv2_dep!=xold_lv2_dep)
	{
		goto STORE;
	}
	// Conserver ces entres valides
	// coup jou prcdemment suivi de la rponse adverse
	// Rutiliser toutes les autres 
}
*/

// Traitement
//#if RESET_HASH_TABLES==1
if (h->key==0) goto STORE;
//#endif
if (h->key!=h_key.key) // On remplace une entre
{
	//if (xhash_indice_generation>h->hash_indice_generation)
	if (xnbcoups!=h->signature_nbcoups)
	{
	    // Copier l'ancienne entre en deuxime partie
		(*(h + xhash_tab_entries_div2)) = (*h);  // copy previous entry to upper half of table
		// Ou craser directement si la ligne prcdente est en commentaire
		goto STORE;
	}
	else if (depth_moins_level>h->depth_moins_level)
	{
		// Copier l'entre en seconde partie	
		h += xhash_tab_entries_div2; 
		goto STORE;
	}
	else // Copier l'entre directement en deuxime partie
	{
		h += xhash_tab_entries_div2; 
		goto STORE;
	}
}

//#if RESET_HASH_TABLES==1
STORE:
//#endif
 xnb_put_hash++;
 h->key = h_key.key;
 h->score = score;
 // Nouvelle version
 // Falphabeta vaut 1,2 ou 3 et est cod sur les 
 // 2 premiers bits
 // next_node_type sur le bit 3
 h->flag=   F_alphabeta
		    #if HASH_TRACK_PV==1
			|(next_node_type==NODE_PV?(1<<2):0)
			#endif
			;
 h->depth_moins_level = depth_moins_level;
 h->move = hmove;
 h->signature_nbcoups=M_128(xnbcoups); 
 //h->hash_indice_generation=xhash_indice_generation;

 //h->lv1_move=xlv1_move;
 //h->lv2_dep=xlv2_dep;
 //h->age=xcurrent_hashtable_age;
//ASSERT(h->age==xcurrent_hashtable_age);

#if DEBUG==1
CONSOLE_MODE;
if (
	1==2
	&&
	h_key.key==11897061026
	&&
	h_key.address==2601402427
	&&
	score==425	
	)
{
printf("\nHash stop");
do 
{
cc=1;
}
while(1);
}
END_MODE;
#endif


// Retour
return;
} // fin put_hash



char get_pawn_hash(H_CODE &h_key)
{
 HASH_PAWN_REC *h;
 char hash;
 STATEVALUES *s;


 // Check lower part of table first then upper part
 h = xhash_pawn_table + (((long)h_key.address)&(xhash_pawn_entries_div2_1));
 if (h->key != (long) h_key.key)
 { h += xhash_pawn_entries_div2; 
    if (h->key != (long) h_key.key) 
		{hash=R_HASH_MISS; return hash;}
	xpawnhash_second_part_hits_this_attempt++;
	#if COLLECT_STATS==1
	xpawnhash_second_part_hits++;
	#endif
 }
 else 
 {
	xpawnhash_first_part_hits_this_attempt++;
	#if COLLECT_STATS==1
	xpawnhash_first_part_hits++;
	#endif
 }

 s=&xstates[xlevel];
 s->pawnstruct=h->pawnstruct;
 s->passed[1]=h->passed[1];
 s->passed[0]=h->passed[0];
 s->passed_maxpos[1]=h->passed_maxpos[1];
 s->passed_maxpos[0]=h->passed_maxpos[0];
 s->colonnes_ouvertes=h->colonnes_ouvertes;
 s->roque[1]=h->roque[1];
 s->roque[0]=h->roque[0];
 s->king_safety[1]=h->king_safety[1];
 s->king_safety[0]=h->king_safety[0];
 return R_HASH_HIT;
}



/* function to find and return a hash entry */
/*int get_hash(H_CODE h_key, int alpha, int beta, int depth,
                           int *hardalpha, int *hardbeta)
*/


char get_hash(H_CODE &h_key,short depth_moins_level,
			  char &hash, MOVE& hmove, short& hscore, char& hscore_status,
			  char &hash_depth_moins_level,
			  char node_type
			  #if HASH_TRACK_PV==1
			  ,char requested_node_type
			  #endif
			  )
{
HASH_REC *h; 
// char cc;
char h_F_alphabeta;
int flag;
char signature_nbcoups;


#if NO_HASH_EXTEND==1
if (xextend) 
{
xhmove=NULL;
hash=R_HASH_MISS; return hash;
}
#endif


#if NO_GET_HASH==1
char dd=1;
if (dd==1) 
{
xhmove=NULL;
hash=R_HASH_MISS; return hash;
}
#endif

if ( !xhash_succes_creation ) 
{
hash=R_HASH_MISS; return hash;
}


// Note : xhscore est retourn  NULL ou avec une valeur

 // Note : conserver variable hash en variable locale
 // En effet dans search
 // Bug !!
 // xhash==R_HASH_MISS => search => xhash=1 => pas d'enregistrement
 // dans la table de la valeur retourne !!

 // Codes retour :
 // 99 - non trouv
 // 1 - trouv
 // 2 - trouv mais avec profondeur insuffisante
 // 3 - upper-bound - lower-bound


 // Vrifier valeur depth
 //if (depth<2*FD) depth=2*FD;


 // Find the right location in the table

 // Attention : il faut un cast sur un long 
 h = xhash_table + (long) (h_key.address&xhash_tab_entries_div2_1);


 // Contrle adresse pointe
#if DEBUG==1
char cc;
cc=0;
HASH_REC *last_hash_table;
last_hash_table=xhash_table+xhash_tab_entries_div2_1;
if (h<xhash_table) cc=1;
if (h>last_hash_table) cc=1;
if (cc==1)
{
log_error("\nErreur H_002");
}
#endif


#if DEBUG==1
CONSOLE_MODE;
if (h_key.address==3175522288 && h_key.key==4175833030)
 {
	 printf("\nHash Trouv 2");
	 getch();
 }
END_MODE;
#endif



 

 // Check lower part of table first then upper part
 if (h->key != (long) h_key.key)
 {  
	h += xhash_tab_entries_div2; 
    if (h->key != (long) h_key.key) 
	{hmove=NULL; hash=R_HASH_MISS; return hash;}
	xhash_second_part_hits_this_attempt++;
	#if COLLECT_STATS==1
	xhash_second_part_hits++;
	#endif
 }
 else 
 {
	xhash_first_part_hits_this_attempt++;
	#if COLLECT_STATS==1
	xhash_first_part_hits++;
	#endif
 }

 
// Mise  jour xhhits2 (nombre de hits non forcment exacts)
#if COLLECT_STATS==1
xhhits2++;
#endif


// Rcupration hash_move
hmove=h->move;
#if NO_RETURN_HMOVE==1
hmove=NULL;
#endif



// Dcomposition flag
flag=h->flag;
h_F_alphabeta=flag&3; // Rcupre les 2 premiers bits
// h_Falphabeta=1 - lower bound 
// h_Falphabeta=2 - window match
// h_Falphabeta=3 - upper bound

// hscore_status prend les valeurs suivantes
// FAIL_LOW, FAIL_HIGH, WINDOW_MATCH, R_HASH_MISS


#if HASH_TRACK_PV==1
if (requested_node_type==NODE_PV)
{
	if ((flag>>2)==0)
	{
		hash_depth_moins_level=-1; // Pour gnrer un FAIL_DEPTH ou 
	}
	#if _DEBUG==1
	cc=0;
	#endif
}
else
#endif
{
	hash_depth_moins_level=h->depth_moins_level;
}
if (hash_depth_moins_level>=depth_moins_level) 
{ 
	hscore_status=h_F_alphabeta; 
	hash=R_HASH_HIT;  
}
else
{
	// Attention : hscore_status ne doit pas retourner
	// R_HASH_FAIL_DEPTH (2) qui serait interprt en WINDOW_MATCH !!
	hscore_status=h_F_alphabeta; 
	hash=R_HASH_FAIL_DEPTH;  
}


// Rcupration score 
hscore = h->score;

// Hash hit ou fail_depth
// R_HASH_FAIL_DEPTH : on laisse hash  fail_depth
// R_HASH_HIT : on transforme en R_HASH_FAIL_BOUNDS
// si on est en dehors des bornes
if (hash==R_HASH_HIT)
{
if (hscore_status!=WINDOW_MATCH)
{
	hash=R_HASH_FAIL_BOUNDS;
}
}

// Ajustement des anciennes valeurs de mat
if (abs(hscore)>MATE_VALUE_THRESHOLD)
{
	signature_nbcoups=M_128(xnbcoups);
	if (signature_nbcoups!=h->signature_nbcoups)
	{
		if (node_type==NODE_PV) 
		{ hash=R_HASH_FAIL_DEPTH; }
		if (hscore<0)
		{
			//hscore+=(xhash_indice_generation!=h->hash_indice_generation?+2:0);
			if (signature_nbcoups-h->signature_nbcoups<20)
			{
				hscore-=signature_nbcoups-h->signature_nbcoups;
				// Attention : on peut avoir trouv un raccourci
				// La position a t enregistre mat en 1  xnbcoups-2
				// et on la redcouvre avec un mat en -1
				// Ceci tant dit, on peut encore annoncer un mat en 1 trop rapidement
				// Cette disposition sert juste  rester en dessous de MATE_VALUE
				if (hscore<-MATE_VALUE) 
				{ 
						hscore=-MATE_VALUE+xlevel-1;
				}
			}
		}
		else //hscore>0
		{
			//hscore+=(xhash_indice_generation!=h->hash_indice_generation?+2:0);
			if (signature_nbcoups-h->signature_nbcoups<20)
			{
				// On peut trouver un raccourci
				// La position a t enregistre mat en 1  xnbcoups-2
				// et on la redcouvre avec un mat annonc en -1
				// Ceci tant dit, on peut encore annoncer un mat en 1 trop rapidement
				// Cette disposition sert juste  rester en dessous de MATE_VALUE
				// En tout tat de cause, il faudrait reconstituer la ligne
				// pour recalculer la valeur de la ligne
				if (hscore>MATE_VALUE) 
				{ 
					hscore=MATE_VALUE-xlevel;
				}
			}
		}
	}
}


/*
// Ajustement des scores de mat
// L'ide est que la valeur mat est stocke  la valeur (mat-ply)+ply
// et recupre  mat-ply+ply-ply soit mat  ply
// C'est indispensable, car simultanment, les valeurs de mat 
// sont diminues du level quand le mat est constat
// Donc un mat constat au level 3 est valu pour MAT-2
// il est enregistr pour MAT+1
// quand il est rcupr ensuite, au coup suivant 
// en hash table au niveau 1, il est rcupr pour MAT+1-1 soit MAT
// Problme : si je rcupre une valeur de mat plus loin
// elle est value pour MAT+ quelque chose ...
// => les valeurs de mat augmentent avec le level
if (hscore>MATE_VALUE_THRESHOLD)
{
	hscore-=xlevel;
}
else if (hscore<-MATE_VALUE_THRESHOLD)
{
	hscore+=xlevel;
}
*/

/*
// Adjust any mate score to be relative to this node.
// if(hscore >= MATE/2) hscore -= ply;
// else if(hscore <= -MATE/2) hscore += ply;
// Si je trouve un mat en 3 au level 6, c'est un mat en 9
if (abs(xhscore)>MATE_VALUE_THRESHOLD)
{
	// Si je trouve 989994 au level 3 (soit mat en 3*2 au level 3)
	// je dois retourner 989992 (mat en 4)
	if (xhscore<0)
	{
		hscore+=xlevel-1; // MAT
	}
	else // xhscore>0
	{
		hscore+=-xlevel+1; // Donne MAT
	}
}
*/

 // return hash hash score if possible, otherwise adjust hardalpha and hardbeta
 /*
 if (h->depth < depth) return hash R_HASH_MISS; 
 if (h->flag==FLAG_A && hscore <= alpha) return hash hscore;
 if (h->flag==FLAG_A && depth > -1) (*hardbeta) = hscore;
 if (h->flag==FLAG_B && hscore >= beta) return hash hscore;
 if (h->flag==FLAG_B && depth > -1) (*hardalpha) = hscore;
 if (h->flag==FLAG_P) return hash hscore;
 return hash R_HASH_MISS;
 */

return hash; 
} // Fin fonction get_hash



/* generate the hash code for a given position */
// also calculate material score
//H_CODE gen_code(position *p)
H_CODE gen_code(bool F_inc)
{

PIECE* wptr;
char i,j;

#if NO_GEN_CODE==1
char dd=1;
if (dd==1) return temp_rec;
#endif
#if DEBUG==1
H_CODE temp_rec2;
#endif

#if INCREMENTAL_HASH==0
bool allow_incremental=false;
H_CODE temp_rec;
#endif
//LEVELVALUES *l2;
char *ep_square64;
#if INCREMENTAL_HASH==1
bool allow_incremental=true;
STATEVALUES *s2;
#endif
H_CODE temp_rec;

// Pourquoi pointer le level+1 ?
// Parce que s->temp_rec est mis  jour par jouer
// au xlevel+1
// ce qui suppose que l'on ait recopi xstates 
// - avant de jouer
// - et avant de calculer gen_code dans tous les cas de figure !!! 
if (xlevel==0) 
{
	F_inc=false;
	#if INCREMENTAL_HASH==1
	s2=&xstates[0];
	#endif
	ep_square64=&xsearch_level[0]->ep_square64;
}
else
{
	#if INCREMENTAL_HASH==1
	s2=&xstates[xlevel+1];
	#endif
	ep_square64=&xsearch_level[xlevel+1]->ep_square64;
}
#if INCREMENTAL_HASH==1
temp_rec=s2->hash_code;
#endif

#if DEBUG==1
if (F_inc)
{
temp_rec2.address=0;
temp_rec2.key=0;
//temp_rec.key2=0;
// Boucle sur les pices blanches
for (j=0; j<=1; j++)
{
wptr=&xw[0+(j<<4)];
for (i=0+(j<<4);i<=xw_size[1-j];i++,wptr++)
{
// Cases vides
if (wptr->nature==0) continue;
or(temp_rec2, wptr->hash_code);
}
}
if (temp_rec2.address!=temp_rec.address || temp_rec2.key != temp_rec.key)
	log_error("\nErr H_656");
}
#endif


// Retour si pas de hash_table
if ( !xhash_succes_creation ) 
{
	temp_rec.address=0;
	temp_rec.key=0;
	//temp_rec.key2=0;
	return temp_rec;
}


// F_inc==false par exemple en reconstitution de ligne principale
// afin de gnrer le bon code
if (!allow_incremental || !F_inc)
{
temp_rec.address=0;
temp_rec.key=0;
//temp_rec.key2=0;
// Boucle sur les pices blanches

for (j=0; j<=1; j++)
{
wptr=&xw[0+(j<<4)];
for (i=0+(j<<4);i<=xw_size[1-j];i++,wptr++)
{
// Cases vides
if (wptr->nature==0) continue;

#if DEBUG==1
char pos64, type;
type=wptr->type;
pos64=wptr->pos;
if (type<0 || type>11 || 
	pos64<0 || pos64>63)
{
log_error("\nPb H_501"); }
#endif

// temp_rec = or(temp_rec, hval[ID(p->sq[i])][i]); 
// temp_rec = or(temp_rec, xhval[wptr->type][wptr->pos64]);
// temp_rec = or(temp_rec, wptr->hash_code);
or(temp_rec, wptr->hash_code);

// Calcul du score matriel et du nombre de pices
//  if(p->sq[i].side == p->wtm) p->material += value[p->sq[i].type];
//    else p->material -= value[p->sq[i].type];
//    if(p->sq[i].type > PAWN) p->pieces[p->sq[i].side]++;

} // fin boucle de hashage // position
} // fin for j


// Reset xstates.hash_code
#if INCREMENTAL_HASH==1
#if DEBUG==1
H_CODE hash_code_mem;
hash_code_mem=s2->hash_code;
#endif
s2->hash_code=temp_rec;
#endif
} // fin test reinitialisation du hashage


or(temp_rec, x_xwtm_castle[xwtm][s2->droit_roque[1]][s2->droit_roque[0]]);
if (*ep_square64!=0)
{
	or(temp_rec, xep_code[*ep_square64]);
}
if (xpasse)
{
	or(temp_rec, xpasse_code[1]);
}


//if   (xwtm==1)  temp_rec = or(temp_rec, xhwtm);
//else            temp_rec = or(temp_rec, xhbtm);
//temp_rec = or(temp_rec, xcastle_code[castle]);
//temp_rec = or(temp_rec, xpasse_code[xpasse]);



/* Ancienne version  conserver
// Hashage trait
//if   (xwtm==1)  temp_rec = or(temp_rec, xhwtm);
//else            temp_rec = or(temp_rec, xhbtm);
if   (xwtm==1)  or(temp_rec, xhwtm);
else            or(temp_rec, xhbtm);
// Question : o btm et wtm sont ils initialiss ?
// Rponse : dans la fonction qui donne le code H

  
// A METTRE A JOUR : la position doit coder le roque
// et la possibilit de faire une prise en passant
castle=		(    statesptr->o_o[xwtm]    
			     + statesptr->o_o_o[xwtm] )
			 // de 0  5
			 +
			(    statesptr->o_o[1-xwtm] 
			     + statesptr->o_o_o[1-xwtm] ) * 6;
			 // de (0  5) * 6

	
#if DEBUG==1
if (castle<0 || castle>35 || 
	Fep<0 || Fep>63 || 
	xwtm<0  || xwtm>1 )
{
log_error("\nPb H_502"); 
}
#endif



//temp_rec = or(temp_rec, xcastle_code[castle]);
//temp_rec = or(temp_rec, xep_code[Fep]);
//temp_rec = or(temp_rec, xpasse_code[xpasse]);

or(temp_rec, xcastle_code[castle]);
or(temp_rec, xep_code[Fep]);
or(temp_rec, xpasse_code[xpasse]);


//  temp_rec = or(temp_rec, xcastle_code[p->castle]);
//  temp_rec = or(temp_rec, xep_code[p->ep]);
*/


#if DEBUG==1 && INCREMENTAL_HASH==1
char cc;
if (F_inc)
{
H_CODE temp_rec2;
temp_rec2=gen_code(0);
if (temp_rec2.address!=temp_rec.address || temp_rec2.key != temp_rec.key)
	log_error("\nErr H_654");
// Cette erreur peut tre lie au faut que l'on a oubli de recopier 
// dans xstates du level suprieur avant de jouer
}
cc=0;
#endif


return temp_rec;
} // fin gen_code


H_CODE gen_code_pawn()
{

char i,j;
char F_presence_dame[2];
STATEVALUES *s;
PIECE *wptr;
bool F_inc;

// Initialisations
s=&xstates[xlevel];
if ( s->materiel_dames[0]==0 ) F_presence_dame[0]=1;
else F_presence_dame[0]=0;
if ( s->materiel_dames[1]==0 ) F_presence_dame[1]=1;
else F_presence_dame[1]=0;
#if INCREMENTAL_HASH==0
bool allow_incremental=false;
H_CODE temp_rec;
#endif
#if INCREMENTAL_HASH==1
bool allow_incremental=true;
H_CODE temp_rec;
//STATEVALUES *s2;
if (xlevel==0) s=&xstates[0];
// Au niveau de xlevel
else s=&xstates[xlevel];
temp_rec=s->pawn_hash_code;
#endif
if (xlevel==0) 
	 F_inc=false; 
else F_inc=true;


#if DEBUG==1
H_CODE temp_rec2;
if (F_inc)
{
temp_rec2.address=0;
temp_rec2.key=0;
//temp_rec.key2=0;
// Boucle sur les pices blanches
for (j=0; j<=1; j++)
{
wptr=&xw[0+(j<<4)];
for (i=0+(j<<4);i<=xw_size[1-j];i++,wptr++)
{
// Cases vides
switch(wptr->nature)
{
case 'P' : 
	or(temp_rec2, wptr->hash_code);
	break;
case 'R' :
	or(temp_rec2, wptr->hash_code);
	break;
default : 
	continue;
}
}
}
if (temp_rec2.address!=temp_rec.address || temp_rec2.key != temp_rec.key)
	log_error("\nErr H_657");
}
#endif


if (!allow_incremental || !F_inc)
{
temp_rec.address=0;
temp_rec.key=0;
// ( Boucle sur les pices )
for (j=0;j<=1;j++)
{
wptr=&xw[0+(j<<4)];
for (i=0+(j<<4);i<=xw_size[1-j];i++,wptr++)
{
// Slections des pions
switch(wptr->nature)
{
	case 'P' :
		or(temp_rec, wptr->hash_code);
		break;
	case 'R' :
		or(temp_rec, wptr->hash_code);
		break;
	default : break;
} 
} // fin boucle de hashage sur les pices
}
// Reset dans xstates
#if INCREMENTAL_HASH==1
s->pawn_hash_code=temp_rec;
#endif
} // fin reinit


// Codage roque
//temp_rec = or(temp_rec, xpawn_roque_code[s->droit_roque[1]+6*s->droit_roque[0]]);
if (s->finale==0)
{
	or(temp_rec, xpawn_roque_code[s->equivalent_droit_roque[1]][s->equivalent_droit_roque[0]]);
}
/*
else
{
	or(temp_rec, xpawn_roque_code[PAS_DE_DROIT_A_ROQUER][PAS_DE_DROIT_A_ROQUER]);
}
*/

// Codage finale et dames 
or(temp_rec, x_finale_et_dames[s->finale][(F_presence_dame[1]<<1)+F_presence_dame[0]]);

/* Ancienne version
// Codage finale
//temp_rec = or(temp_rec, xfinale_code[s->finale]);
or(temp_rec, xfinale_code[s->finale]);

// Codage presence_dames
//temp_rec = or(temp_rec, xpawn_presence_dames[F_presence_dame[1]*2+F_presence_dame[0]]);
or(temp_rec, xpawn_presence_dames[F_presence_dame[1]*2+F_presence_dame[0]]);
*/

#if DEBUG==1
if ( s->finale<0 || s->finale>3 ) 
	log_error("\nErr H_870");
#endif


return temp_rec;
} // fin gen_code_pawn();




// random number generator
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

float ran(long *idum)
{
   int j;
   long k;
   static long iy=0;
   static long iv[NTAB];
   //float temp;
   double temp;

   if (*idum <= 0 || !iy) {
       if (-(*idum) < 1) *idum=1;
       else *idum = -(*idum);
       for (j=NTAB+7; j>=0;j--) {
           k=(*idum)/IQ;
           *idum=IA*(*idum-k*IQ)-IR*k;
           if (*idum < 0) *idum += IM;
           if (j < NTAB) iv[j] = *idum;
       }
       iy=iv[0];
   }
   k=(*idum)/IQ;
   *idum=IA*(*idum-k*IQ)-IR*k;
   if (*idum < 0) *idum += IM;
   j=iy/NDIV;
   iy =iv[j];
   iv[j] = *idum;
   if ((temp=AM*iy) > RNMX) return (float) RNMX;
   else return (float) temp;
}

/* function to return a random bit */
unsigned long ibit(long *idum)
{
  if (ran(idum) > 0.5) return 1;
  return 0;
}

/* function to generate hash codes for all pieces/positions */
void start_code()
{
  long seed1 = -1712913483;
  int i, j, k , l; // l,m


  // Initialisation xhval
  // 12 types de pices diffrents
  for (j = 0; j < 12; j++)
  {
     for (i = 0; i < 64; i++)
      {
       xhval[j][i].address = 0;
       xhval[j][i].key = 0;
       //xhval[j][i].key2 = 0;
       for(k = 0; k < 32; k++)
        {
         xhval[j][i].address = xhval[j][i].address | (ibit(&seed1) << k);
         xhval[j][i].key = xhval[j][i].key | (ibit(&seed1) << k);
         //xhval[j][i].key2 = xhval[j][i].key2 | (ibit(&seed1) << k);
        }
      }
  }

  // Initialisation xwtm_castle
  for (j=0;j<=8;j++)
  {
  for (i=0;i<=8;i++)
  {
  for (l=0;l<=1;l++)
  {
     x_xwtm_castle[l][i][j].address = 0;
     x_xwtm_castle[l][i][j].key = 0;
     //x_xwtm_castle[j][i].key2 = 0;
     for(k = 0; k < 32; k++)
     {
     x_xwtm_castle[l][i][j].address 
		 = x_xwtm_castle[l][i][j].address | (ibit(&seed1) << k);
     x_xwtm_castle[l][i][j].key 
		 = x_xwtm_castle[l][i][j].key | (ibit(&seed1) << k);
     //x_xwtm_castle[j][i].key2 
	 //	 = x_xwtm_castle[j][i].key2 | (ibit(&seed1) << k);
     }
  }}} // Fermetures for


  // Initialisation xep code
  for(i = 0; i < 64; i++) 
  {
    xep_code[i].address = 0;
    xep_code[i].key = 0;
    //xep_code[i].key2 = 0;
    for(k = 0; k < 32; k++)
    {
     xep_code[i].address = xep_code[i].address | (ibit(&seed1) << k);
     xep_code[i].key = xep_code[i].key | (ibit(&seed1) << k);
     //xep_code[i].key2 = xep_code[i].key2 | (ibit(&seed1) << k);
    }
  }

  
  // Initialisation xpasse_code
  // !xpasse => pas de passe
  // xpasse => passe search
  for(i = 0; i < 2; i++) 
  {
    xpasse_code[i].address = 0;
    xpasse_code[i].key = 0;
    //xpasse_code[i].key2 = 0;
    for(k = 0; k < 32; k++) 
    {
     xpasse_code[i].address = xpasse_code[i].address | (ibit(&seed1) << k);
     xpasse_code[i].key = xpasse_code[i].key | (ibit(&seed1) << k);
     //xpasse_code[i].key2 = xpasse_code[i].key2 | (ibit(&seed1) << k);
    }
  }


  /*
  // Initialisation xhwtm et xhbtm
  xhwtm.address = 0; xhwtm.key = 0;
  xhbtm.address = 0; xhbtm.key = 0;
  for(k = 0; k < 32; k++) 
  {
   xhwtm.key = xhwtm.key | (ibit(&seed1) << k);
   xhwtm.address = xhwtm.address | (ibit(&seed1) << k);
   xhbtm.key = xhbtm.key | (ibit(&seed1) << k);
   xhbtm.address = xhbtm.address | (ibit(&seed1) << k);
  }

  // Initialisation castle code
  for(i = 0; i < 36; i++) 
  {
    xcastle_code[i].key = 0;
    xcastle_code[i].address = 0;
    for(k = 0; k < 32; k++) 
    {
     xcastle_code[i].key = xcastle_code[i].key | (ibit(&seed1) << k);
     xcastle_code[i].address = xcastle_code[i].address | (ibit(&seed1) << k);
    }
  }

  // Initialisation ep code
  for(i = 0; i < 64; i++) 
  {
    xep_code[i].key = 0;
    xep_code[i].address = 0;
    for(k = 0; k < 32; k++)
    {
     xep_code[i].key = xep_code[i].key | (ibit(&seed1) << k);
     xep_code[i].address = xep_code[i].address | (ibit(&seed1) << k);
    }
  }

   
   

   // Initialisation xpasse_code
   // !xpasse => pas de passe
   // xpasse => passe search
   for(i = 0; i < 2; i++) 
   {
    xpasse_code[i].key = 0;
    xpasse_code[i].address = 0;
    for(k = 0; k < 32; k++) 
    {
     xpasse_code[i].key = xpasse_code[i].key | (ibit(&seed1) << k);
     xpasse_code[i].address = xpasse_code[i].address | (ibit(&seed1) << k);
    }
   }
   */


   // Initialisation xpawn_roque_code
   // Les valeurs prises sont au nombre de 6
   // (0,1,2,3,4,8)
   // 72=8+16*8
   // Rem *8 est insuffisant si le premier roque vaut 8
   // cela occupe le bit 4 et le dcalage de trois amnerait en 4
   // D'o un dcalage de 4
   for (j = 0; j <= 8; j++)
   {
   for (i = 0; i <= 8; i++)
   {
       xpawn_roque_code[j][i].address = 0;
       xpawn_roque_code[j][i].key = 0;
       //xpawn_roque_code[i].key2 = 0;
	   for(k = 0; k < 32; k++)
	   {
	 	  xpawn_roque_code[j][i].address = xpawn_roque_code[j][i].address | (ibit(&seed1) << k);
		  xpawn_roque_code[j][i].key = xpawn_roque_code[j][i].key | (ibit(&seed1) << k);
		  //xpawn_roque_code[i].key2= xpawn_roque_code[i].key2 | (ibit(&seed1) << k);
	   }
   }
   }
 
   // Initialisation x_finale_et_dames[4][4]
   // Code finale de 
   // De 0  3 ( F_presence_dame[1]*2+F_presence_dame[0] )
   for(i = 0; i < 4; i++) 
   {
   for(j = 0; j < 4; j++) 
   {
       x_finale_et_dames[j][i].address = 0;
       x_finale_et_dames[j][i].key = 0;
       //x_finale_et_dames[j][i].key2 = 0;
       for(k = 0; k < 32; k++)
       {
         x_finale_et_dames[j][i].address = x_finale_et_dames[j][i].address | (ibit(&seed1) << k);
         x_finale_et_dames[j][i].key = x_finale_et_dames[j][i].key | (ibit(&seed1) << k);
         //x_finale_et_dames[j][i].key2 = x_finale_et_dames[j][i].key2 | (ibit(&seed1) << k);
	   }
   }}


   /*
   // Initialisation xpawn_presence_dames
   // De 0  3 ( F_presence_dame[1]*2+F_presence_dame[0] )
   for(i = 0; i <= 3; i++) 
   {
    xpawn_presence_dames[i].key = 0;
    xpawn_presence_dames[i].address = 0;
    for(k = 0; k < 32; k++) 
    {
     xpawn_presence_dames[i].key = xpawn_presence_dames[i].key | (ibit(&seed1) << k);
     xpawn_presence_dames[i].address = xpawn_presence_dames[i].address | (ibit(&seed1) << k);
    }
   }
   


   // Initialisation xfinale_code
   for (i = 0; i <= 3; i++)
   {
       xfinale_code[i].key = 0;
       xfinale_code[i].address = 0;
       for(k = 0; k < 32; k++)
        {
         xfinale_code[i].key = xfinale_code[i].key | (ibit(&seed1) << k);
         xfinale_code[i].address = xfinale_code[i].address | (ibit(&seed1) << k);
        }
   }
   */

}

